#include <iostream>
#include <fstream>
#include <vector>
#include <string>
#include <algorithm>

#include <boost/algorithm/string.hpp>

#include <scip/scip.h>
#include <scip/scipdefplugins.h>
#include <scip/struct_scip.h>

// modified from the scip queens example
#define SCIP_CALL_EXC(x)			\
   {						\
      SCIP_RETCODE retcode;			\
      if( (retcode = (x)) != SCIP_OKAY)		\
      {						\
		throw std::runtime_error("scip returned error code: "  + std::to_string(retcode));		\
      }						\
   }

struct ScipContext 
{
    SCIP* scip;        
    std::vector<SCIP_VAR*> variables;
    std::vector<SCIP_CONS*> constraints;

    ~ScipContext() 
    {
        // because there is no reasonable way to handle SCIP_RETCODE != SCIP_OKAY there is no need to check it.

        // since the SCIPcreateVar captures all variables, we have to release them now
        for (auto var : variables)
        {
            if(var != NULL)
	        {
                SCIPreleaseVar(scip, &var);
            }
        }

        // the same for the constraints
        for (auto constraint : constraints)
        {
	        if(constraint != NULL)
	        {
                SCIPreleaseCons(scip, &constraint);
            }
        }
		
	    // after releasing all vars and cons we can free the scip problem
	    // remember this has allways to be the last call to scip
	    SCIPfree(&scip);
    }
};

class Set 
{
public:
	Set(std::string& line)
	{
		std::vector<std::string> strs;
		boost::split(strs, line, boost::is_any_of(" \r\n"));
		costs = std::stod(strs[0]);
		for(int i = 1; i < strs.size(); ++i)
		{
			if(strs[i] != "")
			{
				members.push_back(std::stoi(strs[i]));
			}
		}
	}

	std::vector<int> members;
	double costs;
};


int main(int argc, char* argv[])
{
	if(argc < 2)
	{
		return 1;
	}

	std::ifstream input(argv[1]);
	std::string line;
	std::vector<Set> sets;
	while (std::getline(input, line))
	{
		sets.emplace_back(line);		
	}

	ScipContext ctx;
	SCIP_CALL_EXC(SCIPcreate(&ctx.scip));
    SCIP_CALL_EXC(SCIPincludeDefaultPlugins(ctx.scip));
    SCIP_CALL_EXC(SCIPcreateProb(ctx.scip, "mip", NULL, NULL, NULL, NULL, NULL, NULL, NULL));

 	ctx.variables.resize(sets.size());
    for (size_t i = 0; i < sets.size(); ++i)
    {
	    SCIP_VAR* x;
	    const auto vartype = SCIP_VARTYPE_INTEGER;
	    const auto varname = "x_" + std::to_string(i);
	    SCIP_CALL_EXC(SCIPcreateVar(ctx.scip, &x, varname.c_str(), 0.0, 1.0, sets[i].costs, vartype, TRUE, FALSE, NULL, NULL, NULL, NULL, NULL));

	    SCIP_CALL_EXC(SCIPaddVar(ctx.scip, x));
	    ctx.variables[i] = x;
    }

    auto linearConstraintCount = 0;
	for(auto& set : sets)
	{
		linearConstraintCount = std::max(linearConstraintCount, 1 + *std::max_element(set.members.begin(), set.members.end()));
	}

    ctx.constraints.resize(linearConstraintCount);

	const auto inf = SCIPinfinity(ctx.scip);
    for (size_t i = 0; i < linearConstraintCount; ++i)
    {
	    SCIP_CONS* cons;
	    const auto name = "c_" + std::to_string(i);

	    SCIP_CALL_EXC(SCIPcreateConsLinear(ctx.scip, &cons, name.c_str(), 0, NULL, NULL, 1.0, inf,
		    TRUE, TRUE, TRUE, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE));

	    ctx.constraints[i] = cons;
    }

    for (size_t col = 0; col < sets.size(); ++col)
    {
	    for (int row : sets[col].members)
	    {
		    SCIP_CALL_EXC(SCIPaddCoefLinear(ctx.scip, ctx.constraints[row], ctx.variables[col], 1.0));
	    }
    }

    for (size_t i = 0; i < ctx.constraints.size(); ++i)
    {
	    SCIP_CALL_EXC(SCIPaddCons(ctx.scip, ctx.constraints[i]));
    }

    SCIP_CALL_EXC(SCIPsolve(ctx.scip));
}
